package pl.edu.amu.wmi.bookapi.service;


import com.google.zxing.BinaryBitmap;
import com.google.zxing.ChecksumException;
import com.google.zxing.FormatException;
import com.google.zxing.LuminanceSource;
import com.google.zxing.NotFoundException;
import com.google.zxing.Reader;
import com.google.zxing.Result;
import com.google.zxing.client.j2se.BufferedImageLuminanceSource;
import com.google.zxing.common.HybridBinarizer;
import com.google.zxing.oned.EAN13Reader;
import org.bytedeco.opencv.global.opencv_imgproc;
import org.bytedeco.opencv.opencv_core.Mat;
import org.bytedeco.opencv.opencv_core.MatVector;
import org.bytedeco.opencv.opencv_core.Size;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.web.multipart.MultipartFile;
import pl.edu.amu.wmi.bookapi.config.ImageProcessingConfigurationProperties;
import pl.edu.amu.wmi.bookapi.exceptions.EanDetectedButUnableToReadException;

import javax.imageio.ImageIO;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.IOException;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.UUID;

import static org.bytedeco.opencv.global.opencv_core.CV_32F;
import static org.bytedeco.opencv.global.opencv_core.convertScaleAbs;
import static org.bytedeco.opencv.global.opencv_core.subtract;
import static org.bytedeco.opencv.global.opencv_imgcodecs.imread;
import static org.bytedeco.opencv.global.opencv_imgcodecs.imwrite;
import static org.bytedeco.opencv.global.opencv_imgproc.Scharr;
import static org.bytedeco.opencv.global.opencv_imgproc.blur;
import static org.bytedeco.opencv.global.opencv_imgproc.contourArea;
import static org.bytedeco.opencv.global.opencv_imgproc.cvtColor;
import static org.bytedeco.opencv.global.opencv_imgproc.findContours;
import static org.bytedeco.opencv.global.opencv_imgproc.getStructuringElement;
import static org.bytedeco.opencv.global.opencv_imgproc.morphologyEx;
import static org.bytedeco.opencv.global.opencv_imgproc.threshold;
import static org.opencv.imgproc.Imgproc.CHAIN_APPROX_SIMPLE;
import static org.opencv.imgproc.Imgproc.COLOR_BGR2GRAY;
import static org.opencv.imgproc.Imgproc.MORPH_CLOSE;
import static org.opencv.imgproc.Imgproc.MORPH_RECT;
import static org.opencv.imgproc.Imgproc.RETR_EXTERNAL;
import static org.opencv.imgproc.Imgproc.THRESH_BINARY;

@Service
public class ImageProcessingService {

    private Logger log = LoggerFactory.getLogger(ImageProcessingService.class);
    private ImageProcessingConfigurationProperties imageProcessingConfigurationProperties;

    private static final Map<String, String> contentTypeConverter = Map.of(
            "image/jpeg", "jpg",
            "image/png", "png");

    private class InvalidTypeException extends RuntimeException {
        public InvalidTypeException(String message) {
            super(message);
        }
    }

    @Autowired
    public ImageProcessingService(ImageProcessingConfigurationProperties imageProcessingConfigurationProperties) {
        this.imageProcessingConfigurationProperties = imageProcessingConfigurationProperties;
    }

    public String getDecodedEan(MultipartFile imageFile) throws Exception {
        if (doesImageContainBarcode(imageFile)){
            return getBarcode(imageFile);
        }
        return "";
    }

    private String getBarcode(MultipartFile imageFile) throws IOException {
        Reader reader = new EAN13Reader();
        BufferedImage img = ImageIO.read(imageFile.getInputStream());
        LuminanceSource source = new BufferedImageLuminanceSource(img);
        BinaryBitmap bitmap = new BinaryBitmap(new HybridBinarizer(source));
        try {
            Result result = reader.decode(bitmap);
            return result.getText();
        } catch (NotFoundException | FormatException | ChecksumException e) {
            e.printStackTrace();
            throw new EanDetectedButUnableToReadException();
        }
    }

    private Boolean doesImageContainBarcode(MultipartFile imageFile) throws IOException{
        String uid = UUID.randomUUID().toString();
        String currentPath = Paths.get(".").toAbsolutePath().normalize().toString();
        String rawContentType = imageFile.getContentType();

        if (!contentTypeConverter.containsKey(rawContentType)) { throw new InvalidTypeException("Only png and jpeg files accepted"); }

        BufferedImage img = ImageIO.read(imageFile.getInputStream());
        String contentType = contentTypeConverter.get(rawContentType);

        String path = currentPath + "/" + uid + "." + contentType;
        File outputfile = new File(path);
        ImageIO.write(img, contentType, outputfile);

        Mat src = imread(path);
        saveDebugImage(src, uid, "initial", contentType);

        Mat dst = new Mat();
        cvtColor(src, dst, COLOR_BGR2GRAY);

        saveDebugImage(dst, uid, "bgr2gray", contentType);

        Mat matScharX = new Mat();
        Mat matScharY = new Mat();

        Scharr(dst, matScharX, CV_32F, 1, 0, 1.0d, -1.0, 0);
        Scharr(dst, matScharY, CV_32F, 0, 1, 1.0d, -1.0, 0);

        subtract(matScharX, matScharY);

        Mat conv = new Mat();
        convertScaleAbs(matScharX, conv);

        saveDebugImage(conv, uid, "scharr", contentType);

        Mat blurDestination = new Mat();
        blur(conv, blurDestination, new Size(9,9));

        saveDebugImage(blurDestination, uid, "blur", contentType);

        Mat thresholdDestination = new Mat();
        threshold(blurDestination, thresholdDestination, 200d, 255d, THRESH_BINARY);

        saveDebugImage(thresholdDestination, uid, "threshold", contentType);

        Mat kernel = getStructuringElement(MORPH_RECT, new Size(21, 7));
        Mat morph = new Mat();
        morphologyEx(thresholdDestination, morph, MORPH_CLOSE, kernel);

        saveDebugImage(morph, uid, "morph", contentType);

        opencv_imgproc.erode(morph, morph, new Mat());
        for(int i = 0 ; i < 2 ; i++) {
            opencv_imgproc.erode(conv, conv, new Mat());
        }

        saveDebugImage(morph, uid, "erode", contentType);

        for(int i = 0 ; i < 2 ; i++) {
            opencv_imgproc.dilate(morph, morph, new Mat());
        }

        saveDebugImage(morph, uid, "dilate", contentType);

        MatVector contours = new MatVector();
        findContours(morph, contours, RETR_EXTERNAL, CHAIN_APPROX_SIMPLE);
        System.out.println("Found " + contours.size() + " contours");

        saveDebugImage(morph, uid, "final", contentType);

        Double totalQualifyingContours = getQualifyingContoursAreas(
                contours,
                img
        ).stream().reduce(0.0, Double::sum);

        System.out.println("Total qualifying contours " + totalQualifyingContours);
        System.out.println("Total area " + img.getWidth() * img.getHeight());
        System.out.println("Minimum area " + img.getWidth() * img.getHeight() * 0.05);
        return totalQualifyingContours > img.getWidth() * img.getHeight() * 0.05;
    }

    private List<Double> getQualifyingContoursAreas(MatVector contours, BufferedImage image) {
        List<Double> contoursAreas = new ArrayList<>();
        System.out.println("Starting contour evaluation, total " + contours.size());
        for(int i = 0 ; i < contours.size() ; i++) {
            Double area = contourArea(contours.get(i));
            if(isContourQualifiable(
                    contourArea(contours.get(i)),
                    (double) image.getWidth() * image.getHeight()
            )) {
                System.out.println("Contour #" +
                        i +
                        " is qualified by area " +
                        contourArea(contours.get(i)) +
                        " - minimum " +
                        image.getHeight() * image.getWidth() * 0.01);
                contoursAreas.add(area);
            } else {
                System.out.println("Contour #" +
                        i +
                        " is NOT qualified by area " +
                        contourArea(contours.get(i)) +
                        " - minimum " +
                        image.getHeight() * image.getWidth() * 0.01);
            }
        }

        return contoursAreas;
    }

    private Boolean isContourQualifiable(Double contourArea, Double imageSize) {
        return contourArea > imageSize * 0.01;
    }

    private void saveDebugImage(Mat imgToWrite, String uid, String operationName, String contentType) {
        if(imageProcessingConfigurationProperties.getDebug()) {
            imwrite("./" + uid + "__" + operationName + "." + contentType, imgToWrite);
        }
    }
}
